
#include "SDL/SDL.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>


/* On my machine at least, record slightly outperforms replay such that
 * for a lot of very fast events it can't quite replay with correct timing.
 * For now we'll work around this by defining a minimum fudge factor such
 * that if the diff is less then that we'll pretend we've hit the event.
 *
 * 10ms seems to work reasonably now.  For the future, we can look at 
 * actually calibrating this or otherwise improving it.
 *
 */
#define MIN_WAIT 10000


#define POLL 0
#define WAIT 1

#include "sdlrecord.h"

char *default_filename=DEFAULT_FILE;
int get_next_event(SDL_Event *rev,int func);
int SDL_PollEvent(SDL_Event *e);
int SDL_WaitEvent(SDL_Event *e);

int SDL_PollEvent(SDL_Event *e) {
    return(get_next_event(e,POLL));
}

int SDL_WaitEvent(SDL_Event *e) {
    return(get_next_event(e,WAIT));
}

int get_next_event(SDL_Event *rev,int func) {
    char *fn; /* filename of input file */
    static int in=0; /* fd of input file */
    static struct timeval evtp; /* time of event */
    static struct timeval lastevtp; /* event time of last event */
    static struct timeval starttp; /* time we started replaying */
    struct timeval currtp; /* current time */
    static struct timeval lasttp; /* time we replayed last event */
    static SDL_Event ev; /* next event to replay */
    static int (*realwe)()=NULL; /* real SDL_WaitEvent function */
    static int (*realpe)()=NULL; /* real SDL_PollEvent function */
    char err_buff[128]; /* a buff for strerror_r */
    static int use_last=0; /* use the cached event rather then falling back */
    static int fallback=0; /* Fall back to real sdl events */
    int evdiff,tmdiff;
    sdlr_header header;

    /* If the input file isn't open (basically, if this is the first time
     * this function is called), open it and read in the header */
    if(!in) {
        if((fn=getenv("SDL_RECORD_PATH"))==NULL) {
            fn=default_filename;
        }
        if((in=open(fn,O_RDONLY))==-1) {
            strerror_r(errno,err_buff,sizeof(err_buff));
            fprintf(stderr,"Can't open input file %s: %s\n",fn,err_buff);
            exit(0);
        }
	if(read(in,&header,sizeof(header))!=sizeof(header)) {
	    fprintf(stderr,"Error reading header\n");
	    exit(1);
	}
	if(header.magic!=MAGIC) {
	    fprintf(stderr,"This doesn't look like a recorded file\n");
	    exit(1);
	}
	if(header.version!=FILEVERSION) {
	    fprintf(stderr,"Wrong file version %d\n",header.version);
	    exit(1);
	}
	gettimeofday(&starttp,NULL);
	lasttp=starttp;
	lastevtp=header.starttime;
    }

    
    if((!fallback)&&((use_last)||((read(in,&evtp,sizeof(evtp))==sizeof(evtp))&&(read(in,&ev,sizeof(SDL_Event))==sizeof(SDL_Event))))) {
	gettimeofday(&currtp,NULL);
	evdiff=( evtp.tv_sec-lastevtp.tv_sec ) * 1000000 + (evtp.tv_usec-lastevtp.tv_usec);
	tmdiff=( currtp.tv_sec-lasttp.tv_sec ) * 1000000 + (currtp.tv_usec-lasttp.tv_usec);

	
	if(evdiff-tmdiff>MIN_WAIT) {
	    if(func==WAIT) {
		/* if we are blocking, usleep. */
		usleep(evdiff-tmdiff-1000);
	    } else {
		/* if we don't have an event to return, return 0 but remember
		 * to save the event we read in for when it's time */
		use_last=1;
		return(0);
	    }
	}  
	
	lastevtp=evtp;
	lasttp=currtp;
	use_last=0;
	*rev=ev;
	return(1);
    } else { 
    /* When we run out of input file, we fall back on real SDL events */
	fallback=1;
	if((!realwe)||(!realpe)) {
	    realwe=get_dlfn("SDL_WaitEvent");
	    realpe=get_dlfn("SDL_PollEvent");
	}
	if(WAIT) {
	    return(realwe(rev));
	} else {
	    return(realpe(rev));
	}
    }
}



